package com.vaadin.addon.beanvalidation;

import java.util.Collection;
import java.util.Locale;

import javax.validation.constraints.NotNull;

import com.vaadin.data.Item;
import com.vaadin.data.util.BeanItem;
import com.vaadin.ui.Field;
import com.vaadin.ui.Form;

/**
 * Vaadin {@link Form} using the JSR-303 (javax.validation) declarative bean
 * validation on its fields.
 * 
 * Most validation is performed using {@link BeanValidationValidator}. In
 * addition, fields are automatically marked as required when necessary based on
 * the {@link NotNull} annotations.
 * 
 * @author Petri Hakala
 * @author Henri Sara
 */
public class BeanValidationForm<T> extends Form {

    private static final long serialVersionUID = 1L;

    private Class<T> beanClass;
    private Locale locale;

    /**
     * Creates a form that performs validation on beans of the type given by
     * beanClass and its subtypes.
     * 
     * @param beanClass
     *            base class of beans for the form
     */
    public BeanValidationForm(Class<T> beanClass) {
        if (beanClass == null) {
            throw new IllegalArgumentException("Bean class cannot be null");
        }
        this.beanClass = beanClass;
    }

    @Override
    public void addField(Object propertyId, Field field) {
        BeanItem<T> item = getItemDataSource();
        field.setPropertyDataSource(item.getItemProperty(propertyId));
        BeanValidationValidator validator = BeanValidationValidator
                .addValidator(field, propertyId, item.getBean().getClass());
        if (locale != null) {
            validator.setLocale(locale);
        }
        super.addField(propertyId, field);
    }

    /**
     * Sets the item data source for the form. The new data source must be of
     * type {@link BeanItem} and its bean must be assignable to the bean class
     * of the form.
     * 
     * {@inheritDoc}
     */
    @Override
    public void setItemDataSource(Item newDataSource) {
        if (!(newDataSource instanceof BeanItem)) {
            throw new IllegalArgumentException("Item must be a BeanItem");
        }
        if (!beanClass.isAssignableFrom(((BeanItem) newDataSource).getBean()
                .getClass())) {
            throw new IllegalArgumentException("Bean must be of type "
                    + beanClass.getName());
        }
        super.setItemDataSource(newDataSource);
    }

    /**
     * Sets the item data source for the form. The new data source must be of
     * type {@link BeanItem} and its bean must be assignable to the bean class
     * of the form.
     * 
     * {@inheritDoc}
     */
    @Override
    public void setItemDataSource(Item newDataSource, Collection propertyIds) {
        if (!(newDataSource instanceof BeanItem)) {
            throw new IllegalArgumentException("Item must be a BeanItem");
        }
        if (!beanClass.isAssignableFrom(((BeanItem) newDataSource).getBean()
                .getClass())) {
            throw new IllegalArgumentException("Bean must be of type "
                    + beanClass.getName());
        }
        super.setItemDataSource(newDataSource, propertyIds);
    }

    @Override
    @SuppressWarnings("unchecked")
    public BeanItem<T> getItemDataSource() {
        return (BeanItem<T>) super.getItemDataSource();
    }

    /**
     * Returns the base class of beans supported by this form.
     * 
     * @return bean class
     */
    public Class<T> getBeanClass() {
        return beanClass;
    }

    /**
     * Sets the locale to be used for validation messages.
     * 
     * This only applies to fields created after setting the locale (i.e. after
     * the next {@link #setItemDataSource()} call), not to the current content
     * of the form.
     */
    @Override
    public void setLocale(Locale locale) {
        this.locale = locale;
    }

    @Override
    public Locale getLocale() {
        return locale;
    }
}